@@ -413,20 +413,28 @@ class BlockRecurrentTransformer(nn.Module):
     @eval_decorator
     def generate(
         self,
-        length,
         prime,
+        length = None,
         xl_memories: List[torch.Tensor] = [],
         states: List[torch.Tensor] = [],
         temperature = 1.,
-        filter_thres = 0.9
+        filter_thres = 0.9,
+        return_memories_and_states = False
     ):
-        orig_len = prime.shape[-1]
+        length = default(length, self.max_seq_len + 1)
+        start_len = prime.shape[-1]
+
+        assert length <= (self.max_seq_len + 1)
+
         output = prime
 
-        for _ in range(length):
+        memories = []
+        states = []
 
-            logits, *_ = self.forward(
-                output[:, -self.max_seq_len:],
+        for ind in range(length - start_len):
+
+            logits, next_memories, next_states = self.forward(
+                output,
                 xl_memories = xl_memories,
                 states = states
             )
@@ -439,7 +447,14 @@ class BlockRecurrentTransformer(nn.Module):
 
             output = torch.cat((output, sampled), dim = -1)
 
-        return output[:, orig_len:]
+            if divisible_by(output.shape[-1] - 1, self.max_seq_len): # on the sampling of the last token in the current window, set new memories and states
+                memories = next_memories
+                states = next_states
+
+        if return_memories_and_states:
+            return output, memories, states
+
+        return output
 
     def forward(
         self,
@@ -551,3 +566,49 @@ class BlockRecurrentTransformer(nn.Module):
         loss = F.cross_entropy(logits, labels, ignore_index = self.ignore_index)
 
         return loss, next_xl_memories, next_states
+
+# recurrent trainer wrapper
+
+@beartype
+class RecurrentTrainerWrapper(nn.Module):
+    def __init__(
+        self,
+        transformer: BlockRecurrentTransformer
+    ):
+        super().__init__()
+        self.transformer = transformer
+        self.seq_len = transformer.max_seq_len
+
+    @eval_decorator
+    @torch.no_grad()
+    def generate(
+        self,
+        prime
+    ):
+        raise NotImplementedError
+
+    def forward(self, x):
+        total_seq_len, seq_len = x.shape[1], self.seq_len
+
+        assert divisible_by(total_seq_len - 1, seq_len), f'length of sequence ({total_seq_len}) must be equal to a multiple of {seq_len} + 1 (one extra token) during training'
+        segments = total_seq_len // seq_len
+
+        total_loss = 0.
+
+        memories = []
+        states = []
+
+        for ind in range(segments):
+            start = ind * seq_len
+            end = start + seq_len + 1
+
+            loss, memories, states = self.transformer(
+                x[:, start:end],
+                xl_memories = memories,
+                states = states,
+                return_loss = True
+            )
+
+            total_loss = total_loss + (loss / segments)
+
+        return total_loss
